@@ -21,12 +21,14 @@
 
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 
+#include "util.h"
 #include "zlibsupport.h"
 #include "depmod.h"
 #include "logging.h"
 #include "index.h"
-#include "moduleops.h"
+#include "elfops.h"
 #include "tables.h"
+#include "config_filter.h"
 
 #include "testing.h"
 
@@ -54,16 +56,19 @@ struct module_search
 
 	/* search path */
 	char *search_path;
+	size_t len;
 };
 
-static int verbose;
 static unsigned int skipchars;
+static unsigned int make_map_files = 1; /* default to on */
+static unsigned int force_map_files = 0; /* default to on */
 
 #define SYMBOL_HASH_SIZE 1024
 struct symbol
 {
 	struct symbol *next;
 	struct module *owner;
+	uint64_t ver;
 	char name[0];
 };
 
@@ -82,12 +87,13 @@ static inline unsigned int tdb_hash(const char *name)
 	return (1103515243 * value + 12345);
 }
 
-void add_symbol(const char *name, struct module *owner)
+void add_symbol(const char *name, uint64_t ver, struct module *owner)
 {
 	unsigned int hash;
 	struct symbol *new = NOFAIL(malloc(sizeof *new + strlen(name) + 1));
 
 	new->owner = owner;
+	new->ver = ver;
 	strcpy(new->name, name);
 
 	hash = tdb_hash(name) % SYMBOL_HASH_SIZE;
@@ -95,9 +101,10 @@ void add_symbol(const char *name, struct module *owner)
 	symbolhash[hash] = new;
 }
 
-static int print_unknown;
+static int print_unknown, check_symvers;
 
-struct module *find_symbol(const char *name, const char *modname, int weak)
+struct module *find_symbol(const char *name, uint64_t ver,
+		const char *modname, int weak)
 {
 	struct symbol *s;
 
@@ -107,7 +114,13 @@ struct module *find_symbol(const char *name, const char *modname, int weak)
 
 	for (s = symbolhash[tdb_hash(name) % SYMBOL_HASH_SIZE]; s; s=s->next) {
 		if (streq(s->name, name))
-			return s->owner;
+			break;
+	}
+	if (s) {
+		if (ver && s->ver && s->ver != ver && print_unknown && !weak)
+			warn("%s disagrees about version of symbol %s\n",
+					modname, name);
+		return s->owner;
 	}
 
 	if (print_unknown && !weak)
@@ -128,6 +141,14 @@ void add_dep(struct module *mod, struct module *depends_on)
 	mod->deps[mod->num_deps++] = depends_on;
 }
 
+static void add_fake_syms(void)
+{
+	/* __this_module is magic inserted by kernel loader. */
+	add_symbol("__this_module", 0, NULL);
+	/* On S390, this is faked up too */
+	add_symbol("_GLOBAL_OFFSET_TABLE_", 0, NULL);
+}
+
 static void load_system_map(const char *filename)
 {
 	FILE *system_map;
@@ -153,33 +174,58 @@ static void load_system_map(const char *filename)
 			continue;
 
 		/* Covers gpl-only and normal symbols. */
-		if (strncmp(ptr+1, ksymstr, ksymstr_len) == 0)
-			add_symbol(ptr+1+ksymstr_len, NULL);
+		if (strstarts(ptr+1, ksymstr))
+			add_symbol(ptr+1+ksymstr_len, 0, NULL);
 	}
 
 	fclose(system_map);
+	add_fake_syms();
+}
 
-	/* __this_module is magic inserted by kernel loader. */
-	add_symbol("__this_module", NULL);
-	/* On S390, this is faked up too */
-	add_symbol("_GLOBAL_OFFSET_TABLE_", NULL);
+static void load_module_symvers(const char *filename)
+{
+	FILE *module_symvers;
+	char line[10240];
+
+	module_symvers = fopen(filename, "r");
+	if (!module_symvers)
+		fatal("Could not open '%s': %s\n", filename, strerror(errno));
+
+	/* eg. "0xb352177e\tfind_first_bit\tvmlinux\tEXPORT_SYMBOL" */
+	while (fgets(line, sizeof(line)-1, module_symvers)) {
+		const char *ver, *sym, *where;
+
+		ver = strtok(line, " \t");
+		sym = strtok(NULL, " \t");
+		where = strtok(NULL, " \t");
+		if (!ver || !sym || !where)
+			continue;
+
+		if (streq(where, "vmlinux"))
+			add_symbol(sym, strtoull(ver, NULL, 16), NULL);
+	}
+
+	fclose(module_symvers);
+	add_fake_syms();
 }
 
 static struct option options[] = { { "all", 0, NULL, 'a' },
 				   { "quick", 0, NULL, 'A' },
 				   { "basedir", 1, NULL, 'b' },
-				   { "errsyms", 0, NULL, 'e' },
+				   { "config", 1, NULL, 'C' },
+				   { "symvers", 1, NULL, 'E' },
 				   { "filesyms", 1, NULL, 'F' },
-				   { "help", 0, NULL, 'h' },
-				   { "show", 0, NULL, 'n' },
-				   { "dry-run", 0, NULL, 'n' },
+				   { "errsyms", 0, NULL, 'e' },
+				   { "unresolved-error", 0, NULL, 'u' },
 				   { "quiet", 0, NULL, 'q' },
 				   { "root", 0, NULL, 'r' },
-				   { "unresolved-error", 0, NULL, 'u' },
 				   { "verbose", 0, NULL, 'v' },
+				   { "show", 0, NULL, 'n' },
+				   { "dry-run", 0, NULL, 'n' },
+				   { "help", 0, NULL, 'h' },
 				   { "version", 0, NULL, 'V' },
-				   { "config", 1, NULL, 'C' },
-				   { "warn", 1, NULL, 'w' },
+				   { "warn", 0, NULL, 'w' },
+				   { "map", 0, NULL, 'm' },
 				   { NULL, 0, NULL, 0 } };
 
 /* Version number or module name?  Don't assume extension. */
@@ -209,44 +255,11 @@ static int old_module_version(const char *version)
 	return 1;
 }
 
-static void exec_old_depmod(char *argv[])
-{
-	char *sep;
-	char pathname[strlen(argv[0])+1];
-	char oldname[strlen("depmod") + strlen(argv[0]) + sizeof(".old")];
-
-	memset(pathname, 0, strlen(argv[0])+1);
-	sep = strrchr(argv[0], '/');
-	if (sep)
-		memcpy(pathname, argv[0], sep - argv[0]+1);
-	sprintf(oldname, "%s%s.old", pathname, "depmod");
-
-	/* Recursion detection: we need an env var since we can't
-	   change argv[0] (as older modutils uses it to determine
-	   behavior). */
-	if (getenv("MODULE_RECURSE"))
-		return;
-	setenv("MODULE_RECURSE", "y", 0);
-
-	execvp(oldname, argv);
-	fprintf(stderr,
-		"Version requires old depmod, but couldn't run %s: %s\n",
-		oldname, strerror(errno));
-	exit(2);
-}
-
-static void grammar(const char *cmd, const char *filename, unsigned int line)
-{
-	warn("%s line %u: ignoring bad line starting with '%s'\n",
-						filename, line, cmd);
-}
-
-
 static void print_usage(const char *name)
 {
 	fprintf(stderr,
 	"%s " VERSION " -- part of " PACKAGE "\n"
-	"%s -[aA] [-n -e -v -q -V -r -u -w]\n"
+	"%s -[aA] [-n -e -v -q -V -r -u -w -m]\n"
 	"      [-b basedirectory] [forced_version]\n"
 	"depmod [-n -e -v -q -r -u -w] [-F kernelsyms] module1.ko module2.ko ...\n"
 	"If no arguments (except options) are given, \"depmod -a\" is assumed\n"
@@ -257,8 +270,9 @@ static void print_usage(const char *name)
 	"Options:\n"
 	"\t-a, --all            Probe all modules\n"
 	"\t-A, --quick          Only does the work if there's a new module\n"
-	"\t-n, --show           Write the dependency file on stdout only\n"
 	"\t-e, --errsyms        Report not supplied symbols\n"
+	"\t-m, --map            Create the legacy map files\n"
+	"\t-n, --show           Write the dependency file on stdout only\n"
 	"\t-V, --version        Print the release version\n"
 	"\t-v, --verbose        Enable verbose mode\n"
 	"\t-w, --warn		Warn on duplicates\n"
@@ -269,7 +283,10 @@ static void print_usage(const char *name)
 	"\t    --basedir basedirectory    Use an image of a module tree.\n"
 	"\t-F kernelsyms\n"
 	"\t    --filesyms kernelsyms      Use the file instead of the\n"
-	"\t                               current kernel symbols.\n",
+	"\t                               current kernel symbols.\n"
+	"\t-E Module.symvers\n"
+	"\t    --symvers Module.symvers   Use Module.symvers file to check\n"
+	"\t                               symbol versions.\n",
 	"depmod", "depmod");
 }
 
@@ -288,25 +305,6 @@ static int ends_in(const char *name, const char *ext)
 	return 0;
 }
 
-/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
-int needconv(const char *elfhdr)
-{
-	union { short s; char c[2]; } endian_test;
-
-	endian_test.s = 1;
-	if (endian_test.c[1] == 1) return elfhdr[EI_DATA] != ELFDATA2MSB;
-	if (endian_test.c[0] == 1) return elfhdr[EI_DATA] != ELFDATA2LSB;
-	else
-		abort();
-}
-
-static char *my_basename(const char *name)
-{
-	const char *base = strrchr(name, '/');
-	if (base) return (char *)base + 1;
-	return (char *)name;
-}
-
 static struct module *grab_module(const char *dirname, const char *filename)
 {
 	struct module *new;
@@ -320,40 +318,16 @@ static struct module *grab_module(const char *dirname, const char *filename)
 	new->basename = my_basename(new->pathname);
 
 	INIT_LIST_HEAD(&new->dep_list);
+	new->order = INDEX_PRIORITY_MIN;
 
-	new->data = grab_file(new->pathname, &new->len);
-	if (!new->data) {
+	new->file = grab_elf_file(new->pathname);
+	if (!new->file) {
 		warn("Can't read module %s: %s\n",
 		     new->pathname, strerror(errno));
-		goto fail_data;
+		free(new);
+		return NULL;
 	}
-
-	/* "\177ELF" <byte> where byte = 001 for 32-bit, 002 for 64 */
-	if (memcmp(new->data, ELFMAG, SELFMAG) != 0) {
-		warn("Module %s is not an elf object\n", new->pathname);
-		goto fail;
-	}
-
-	switch (((char *)new->data)[EI_CLASS]) {
-	case ELFCLASS32:
-		new->ops = &mod_ops32;
-		break;
-	case ELFCLASS64:
-		new->ops = &mod_ops64;
-		break;
-	default:
-		warn("Module %s has elf unknown identifier %i\n",
-		     new->pathname, ((char *)new->data)[EI_CLASS]);
-		goto fail;
-	}
-	new->conv = needconv(new->data);
 	return new;
-
-fail:
-	release_file(new->data, new->len);
-fail_data:
-	free(new);
-	return NULL;
 }
 
 struct module_traverse
@@ -450,30 +424,18 @@ static void del_module(struct module **modules, struct module *delme)
 	deleted = delme;
 }
 
-/* Convert filename to the module name.  Works if filename == modname, too. */
-static void filename2modname(char *modname, const char *filename)
+/* convert to relative path if possible */
+static const char *compress_path(const char *path, const char *basedir)
 {
-	const char *afterslash;
-	unsigned int i;
+	int len = strlen(basedir);
 
-	afterslash = strrchr(filename, '/');
-	if (!afterslash)
-		afterslash = filename;
-	else
-		afterslash++;
-
-	/* Convert to underscores, stop at first . */
-	for (i = 0; afterslash[i] && afterslash[i] != '.'; i++) {
-		if (afterslash[i] == '-')
-			modname[i] = '_';
-		else
-			modname[i] = afterslash[i];
-	}
-	modname[i] = '\0';
+	if (strncmp(path, basedir, len) == 0)
+		path += len + 1;
+	return path;
 }
 
 static void output_deps(struct module *modules,
-			FILE *out)
+			FILE *out, char *dirname)
 {
 	struct module *i;
 
@@ -481,11 +443,12 @@ static void output_deps(struct module *modules,
 		struct list_head *j, *tmp;
 		order_dep_list(i, i);
 
-		fprintf(out, "%s:", i->pathname + skipchars);
+		fprintf(out, "%s:", compress_path(i->pathname, dirname));
 		list_for_each_safe(j, tmp, &i->dep_list) {
 			struct module *dep
 				= list_entry(j, struct module, dep_list);
-			fprintf(out, " %s", dep->pathname + skipchars);
+			fprintf(out, " %s",
+			        compress_path(dep->pathname, dirname));
 			list_del_init(j);
 		}
 		fprintf(out, "\n");
@@ -496,7 +459,7 @@ static void output_deps(struct module *modules,
 int warn_dups = 0;
 
 static void output_deps_bin(struct module *modules,
-			FILE *out)
+			FILE *out, char *dirname)
 {
 	struct module *i;
 	struct index_node *index;
@@ -511,18 +474,21 @@ static void output_deps_bin(struct module *modules,
 		
 		order_dep_list(i, i);
 		
-		filename2modname(modname, i->pathname + skipchars);
-		asprintf(&line, "%s %s:", modname, i->pathname + skipchars);
+		filename2modname(modname, i->pathname);
+		nofail_asprintf(&line, "%s:",
+		                compress_path(i->pathname, dirname));
 		p = line;
 		list_for_each_safe(j, tmp, &i->dep_list) {
 			struct module *dep
 				= list_entry(j, struct module, dep_list);
-			asprintf(&line, "%s %s", p, dep->pathname + skipchars);
+			nofail_asprintf(&line, "%s %s",
+			                p,
+			                compress_path(dep->pathname, dirname));
 			free(p);
 			p = line;
 			list_del_init(j);
 		}
-		if (index_insert(index, line) && warn_dups)
+		if (index_insert(index, modname, line, i->order) && warn_dups)
 			warn("duplicate module deps:\n%s\n",line);
 		free(line);
 	}
@@ -547,7 +513,6 @@ static int is_higher_priority(const char *newpath, const char *oldpath,
 			      struct module_search *search,
 			      struct module_overrides *overrides)
 {
-	char *p, *q, *s;
 	struct module_search *tmp;
 	struct module_overrides *ovtmp;
 	int i = 0;
@@ -555,63 +520,29 @@ static int is_higher_priority(const char *newpath, const char *oldpath,
 	int prio_new = -1;
 	int prio_old = -1;
 
-/* The names already match, now we check for wildcard overrides and other
- * high priority overrides we added to the config.
+/* The names already match, now we check for overrides and directory search
+ * order
  */
-	for (ovtmp=overrides;ovtmp!=NULL;ovtmp=ovtmp->next) {
-
-		p = strstr(ovtmp->modfile,newpath);
-		if (p)
+	for (ovtmp = overrides; ovtmp != NULL; ovtmp = ovtmp->next) {
+		if (streq(ovtmp->modfile, newpath))
 			return 1;
-
-		q = strstr(ovtmp->modfile,"*");
-		if (q) {
-			p = strstr(newpath, q+1);
-			if (p)
-				return 1;
-
-			p = strstr(oldpath, q+1);
-			if (p)
-				return 0;
-		}
-
-		p = strstr(ovtmp->modfile,oldpath);
-		if (p)
+		if (streq(ovtmp->modfile, oldpath))
 			return 0;
 	}
-	
-	for (i=0,tmp=search;tmp!=NULL;tmp=tmp->next,i++) {
-
-		s = NOFAIL(malloc(strlen(tmp->search_path)+2));
-		s[0] = '/';
-		strncpy(s+1,tmp->search_path, strlen(tmp->search_path)+1);
-
-		if (0 == strncmp(tmp->search_path,MODULE_BUILTIN_KEY,
-				 strlen(MODULE_BUILTIN_KEY)))
+	for (i = 0, tmp = search; tmp != NULL; tmp = tmp->next, i++) {
+		if (streq(tmp->search_path, MODULE_BUILTIN_KEY))
 			prio_builtin = i;
-			
-		p = strstr(newpath,s);
-		if ((p) && ((p[strlen(s)] == '/')
-			||  (p[strlen(s)] == '\0')))
+		else if (strncmp(tmp->search_path, newpath, tmp->len) == 0)
 			prio_new = i;
-
-		p = strstr(oldpath,s);
-		if ((p) && ((p[strlen(s)] == '/')
-			||  (p[strlen(s)] == '\0')))
+		else if (strncmp(tmp->search_path, oldpath, tmp->len) == 0)
 			prio_old = i;
-	
-		free(s);
-		
 	}
-
 	if (prio_new < 0)
 		prio_new = prio_builtin;
-	
 	if (prio_old < 0)
 		prio_old = prio_builtin;
 
 	return prio_new > prio_old;
-	
 }
 
 
@@ -706,17 +637,122 @@ static struct module *grab_basedir(const char *dirname,
 	return list;
 }
 
+static struct module *sort_modules(const char *dirname, struct module *list)
+{
+	struct module *tlist = NULL, **tpos = &tlist;
+	FILE *modorder;
+	int dir_len = strlen(dirname) + 1;
+	char file_name[dir_len + strlen("modules.order") + 1];
+	char line[10240];
+	unsigned int linenum = 0;
+
+	sprintf(file_name, "%s/%s", dirname, "modules.order");
+
+	modorder = fopen(file_name, "r");
+	if (!modorder) {
+		/* Older kernels don't generate modules.order.  Just
+		   return if the file doesn't exist. */
+		if (errno == ENOENT)
+			return list;
+		fatal("Could not open '%s': %s\n", file_name, strerror(errno));
+	}
+
+	sprintf(line, "%s/", dirname);
+
+	/* move modules listed in modorder file to tlist in order */
+	while (fgets(line, sizeof(line), modorder)) {
+		struct module **pos, *mod;
+		int len = strlen(line);
+
+		linenum++;
+		if (line[len - 1] == '\n')
+			line[len - 1] = '\0';
+
+		for (pos = &list; (mod = *pos); pos = &(*pos)->next) {
+			if (streq(line, mod->pathname + dir_len)) {
+				mod->order = linenum;
+				*pos = mod->next;
+				mod->next = NULL;
+				*tpos = mod;
+				tpos = &mod->next;
+				break;
+			}
+		}
+	}
+
+	/* append the rest */
+	*tpos = list;
+
+	fclose(modorder);
+
+	return tlist;
+}
+
+/* Calculate the dependencies for this module */
+static void calculate_deps(struct module *module)
+{
+	unsigned int i;
+	struct string_table *symnames;
+	struct string_table *symtypes;
+	uint64_t *symvers = NULL;
+	struct elf_file *file;
+
+	module->num_deps = 0;
+	module->deps = NULL;
+	file = module->file;
+
+	symnames = file->ops->load_dep_syms(file, &symtypes,
+			check_symvers ? &symvers : NULL);
+	if (!symnames || !symtypes)
+		return;
+
+	for (i = 0; i < symnames->cnt; i++) {
+		const char *name;
+		uint64_t ver;
+		struct module *owner;
+		int weak;
+
+		name = symnames->str[i];
+		ver = symvers ? symvers[i] : 0;
+		weak = (*(symtypes->str[i]) == 'W');
+		owner = find_symbol(name, ver, module->pathname, weak);
+		if (owner) {
+			info("%s needs \"%s\": %s\n",
+			       module->pathname, name,
+			       owner->pathname);
+			add_dep(module, owner);
+		}
+	}
+
+	free(symnames);
+	free(symtypes);
+	free(symvers);
+}
+
 static struct module *parse_modules(struct module *list)
 {
 	struct module *i;
+	struct elf_file *file;
+	struct string_table *syms;
+	int j;
 
 	for (i = list; i; i = i->next) {
-		i->ops->load_symbols(i);
-		i->ops->fetch_tables(i);
+		uint64_t *symvers = NULL;
+		file = i->file;
+		syms = file->ops->load_symbols(file,
+				check_symvers ? &symvers : NULL);
+		if (syms) {
+			for (j = 0; j < syms->cnt; j++)
+				add_symbol(syms->str[j],
+					symvers ? symvers[j] : 0, i);
+			strtbl_free(syms);
+		}
+		free(symvers);
+		file->ops->fetch_tables(file, &i->tables);
 	}
 	
 	for (i = list; i; i = i->next)
-		i->ops->calculate_deps(i, verbose);
+		calculate_deps(i);
 	
 	/* Strip out modules with dependency loops. */
  again:
@@ -733,7 +769,7 @@ static struct module *parse_modules(struct module *list)
 }
 
 /* Simply dump hash table. */
-static void output_symbols(struct module *unused, FILE *out)
+static void output_symbols(struct module *unused, FILE *out, char *dirname)
 {
 	unsigned int i;
 
@@ -752,11 +788,12 @@ static void output_symbols(struct module *unused, FILE *out)
 	}
 }
 
-static void output_symbols_bin(struct module *unused, FILE *out)
+static void output_symbols_bin(struct module *unused, FILE *out, char *dirname)
 {
 	struct index_node *index;
 	unsigned int i;
-	char *line;
+	char *alias;
+	int duplicate;
 
 	index = index_create();
 	
@@ -767,12 +804,13 @@ static void output_symbols_bin(struct module *unused, FILE *out)
 			if (s->owner) {
 				char modname[strlen(s->owner->pathname)+1];
 				filename2modname(modname, s->owner->pathname);
-				asprintf(&line, "symbol:%s %s",
-					s->name, modname);
-				if (index_insert(index, line) && warn_dups)
-					warn("duplicate module syms:\n%s\n",
-						line);
-				free(line);
+				nofail_asprintf(&alias, "symbol:%s", s->name);
+				duplicate = index_insert(index, alias, modname,
+							 s->owner->order);
+				if (duplicate && warn_dups)
+					warn("duplicate module syms:\n%s %s\n",
+						alias, modname);
+				free(alias);
 			}
 		}
 	}
@@ -781,116 +819,84 @@ static void output_symbols_bin(struct module *unused, FILE *out)
 	index_destroy(index);
 }
 
-static const char *next_string(const char *string, unsigned long *secsize)
-{
-	/* Skip non-zero chars */
-	while (string[0]) {
-		string++;
-		if ((*secsize)-- <= 1)
-			return NULL;
-	}
-
-	/* Skip any zero padding. */
-	while (!string[0]) {
-		string++;
-		if ((*secsize)-- <= 1)
-			return NULL;
-	}
-	return string;
-}
-
-/* Careful!  Don't munge - in [ ] as per Debian Bug#350915 */
-static char *underscores(char *string)
-{
-	unsigned int i;
-
-	if (!string)
-		return NULL;
-		
-	for (i = 0; string[i]; i++) {
-		switch (string[i]) {
-		case '-':
-			string[i] = '_';
-			break;
-		
-		case '[':
-			i += strcspn(&string[i], "]");
-			if (!string[i])
-				warn("Unmatched bracket in %s\n", string);
-		}
-	}
-	return string;
-}
-
-static void output_aliases(struct module *modules, FILE *out)
+static void output_aliases(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
-	const char *p;
-	unsigned long size;
+	struct elf_file *file;
+	struct string_table *tbl;
+	int j;
 
 	fprintf(out, "# Aliases extracted from modules themselves.\n");
 	for (i = modules; i; i = i->next) {
 		char modname[strlen(i->pathname)+1];
 
+		file = i->file;
 		filename2modname(modname, i->pathname);
 
 		/* Grab from old-style .modalias section. */
-		for (p = i->ops->get_aliases(i, &size);
-		     p;
-		     p = next_string(p, &size))
-			fprintf(out, "alias %s %s\n", p, modname);
+		tbl = file->ops->load_strings(file, ".modalias", NULL, fatal);
+		for (j = 0; tbl && j < tbl->cnt; j++)
+			fprintf(out, "alias %s %s\n", tbl->str[j], modname);
+		strtbl_free(tbl);
 
-		/* Grab form new-style .modinfo section. */
-		for (p = i->ops->get_modinfo(i, &size);
-		     p;
-		     p = next_string(p, &size)) {
-			if (strncmp(p, "alias=", strlen("alias=")) == 0)
+		/* Grab from new-style .modinfo section. */
+		tbl = file->ops->load_strings(file, ".modinfo", NULL, fatal);
+		for (j = 0; tbl && j < tbl->cnt; j++) {
+			const char *p = tbl->str[j];
+			if (strstarts(p, "alias="))
 				fprintf(out, "alias %s %s\n",
 					p + strlen("alias="), modname);
 		}
+		strtbl_free(tbl);
 	}
 }
 
-static void output_aliases_bin(struct module *modules, FILE *out)
+static void output_aliases_bin(struct module *modules, FILE *out, char *dirname)
 {
 	struct module *i;
-	const char *p;
-	char *line;
-	unsigned long size;
+	struct elf_file *file;
+	struct string_table *tbl;
+	int j;
+	char *alias;
 	struct index_node *index;
+	int duplicate;
 
 	index = index_create();
 	
 	for (i = modules; i; i = i->next) {
 		char modname[strlen(i->pathname)+1];
 
+		file = i->file;
 		filename2modname(modname, i->pathname);
 
 		/* Grab from old-style .modalias section. */
-		for (p = i->ops->get_aliases(i, &size);
-		     p;
-		     p = next_string(p, &size)) {
-			asprintf(&line, "%s %s", p, modname);
-			underscores(line);
-			if (index_insert(index, line) && warn_dups)
-				warn("duplicate module alias:\n%s\n", line);
-			free(line);
+		tbl = file->ops->load_strings(file, ".modalias", NULL, fatal);
+		for (j = 0; tbl && j < tbl->cnt; j++) {
+			alias = NOFAIL(strdup(tbl->str[j]));
+			underscores(alias);
+			duplicate = index_insert(index, alias, modname, i->order);
+			if (duplicate && warn_dups)
+				warn("duplicate module alias:\n%s %s\n",
+					alias, modname);
+			free(alias);
 		}
+		strtbl_free(tbl);
 
 		/* Grab from new-style .modinfo section. */
-		for (p = i->ops->get_modinfo(i, &size);
-		     p;
-		     p = next_string(p, &size)) {
-			if (strncmp(p, "alias=", strlen("alias=")) == 0) {
-				asprintf(&line, "%s %s",
-					p + strlen("alias="), modname);
-				underscores(line);
-				if (index_insert(index, line) && warn_dups)
-					warn("duplicate module alias:\n%s\n",
-						line);
-				free(line);
+		tbl = file->ops->load_strings(file, ".modinfo", NULL, fatal);
+		for (j = 0; tbl && j < tbl->cnt; j++) {
+			const char *p = tbl->str[j];
+			if (strstarts(p, "alias=")) {
+				alias = NOFAIL(strdup(p + strlen("alias=")));
+				underscores(alias);
+				duplicate = index_insert(index, alias, modname, i->order);
+				if (duplicate && warn_dups)
+					warn("duplicate module alias:\n%s %s\n",
+						alias, modname);
+				free(alias);
 			}
 		}
+		strtbl_free(tbl);
 	}
 	
 	index_write(index, out);
@@ -899,24 +905,25 @@ static void output_aliases_bin(struct module *modules, FILE *out)
 
 struct depfile {
 	char *name;
-	void (*func)(struct module *, FILE *);
+	void (*func)(struct module *, FILE *, char *dirname);
+	int map_file;
 };
 
 static struct depfile depfiles[] = {
-	{ "modules.dep", output_deps }, /* This is what we check for '-A'. */
-	{ "modules.dep.bin", output_deps_bin },
-	{ "modules.pcimap", output_pci_table },
-	{ "modules.usbmap", output_usb_table },
-	{ "modules.ccwmap", output_ccw_table },
-	{ "modules.ieee1394map", output_ieee1394_table },
-	{ "modules.isapnpmap", output_isapnp_table },
-	{ "modules.inputmap", output_input_table },
-	{ "modules.ofmap", output_of_table },
-	{ "modules.seriomap", output_serio_table },
-	{ "modules.alias", output_aliases },
-	{ "modules.alias.bin", output_aliases_bin },
-	{ "modules.symbols", output_symbols },
-	{ "modules.symbols.bin", output_symbols_bin }
+	{ "modules.dep", output_deps, 0 }, /* This is what we check for '-A'. */
+	{ "modules.dep.bin", output_deps_bin, 0 },
+	{ "modules.pcimap", output_pci_table, 1 },
+	{ "modules.usbmap", output_usb_table, 1 },
+	{ "modules.ccwmap", output_ccw_table, 1 },
+	{ "modules.ieee1394map", output_ieee1394_table, 1 },
+	{ "modules.isapnpmap", output_isapnp_table, 1 },
+	{ "modules.inputmap", output_input_table, 1 },
+	{ "modules.ofmap", output_of_table, 1 },
+	{ "modules.seriomap", output_serio_table, 1 },
+	{ "modules.alias", output_aliases, 0 },
+	{ "modules.alias.bin", output_aliases_bin, 0 },
+	{ "modules.symbols", output_symbols, 0 },
+	{ "modules.symbols.bin", output_symbols_bin, 0 }
 };
 
 /* If we can't figure it out, it's safe to say "true". */
@@ -969,52 +976,6 @@ static int depfile_out_of_date(const char *dirname)
 	return any_modules_newer(dirname, st.st_mtime);
 }
 
-static char *getline_wrapped(FILE *file, unsigned int *linenum)
-{
-	int size = 256;
-	int i = 0;
-	char *buf = NOFAIL(malloc(size));
-	for(;;) {
-		int ch = getc_unlocked(file);
-		
-		switch(ch) {
-		case EOF:
-			if (i == 0) {
-				free(buf);
-				return NULL;
-			}
-			/* else fall through */
-			
-		case '\n':
-			if (linenum)
-				(*linenum)++;
-			if (i == size)
-				buf = NOFAIL(realloc(buf, size + 1));
-			buf[i] = '\0';
-			return buf;
-			
-		case '\\':
-			ch = getc_unlocked(file);
-			
-			if (ch == '\n') {
-				if (linenum)
-					(*linenum)++;
-				continue;
-			}
-			/* else fall through */
-		
-		default:
-			buf[i++] = ch;
-	
-			if (i == size) {
-				size *= 2;
-				buf = NOFAIL(realloc(buf, size));
-			}
-		}
-	}
-}
-
-
 static char *strsep_skipspace(char **string, char *delim)
 {
 	if (!*string)
@@ -1024,6 +985,7 @@ static char *strsep_skipspace(char **string, char *delim)
 }
 
 static struct module_search *add_search(const char *search_path,
+					size_t len,
 					struct module_search *search)
 {
 
@@ -1031,6 +993,7 @@ static struct module_search *add_search(const char *search_path,
 	
 	new = NOFAIL(malloc(sizeof(*new)));
 	new->search_path = NOFAIL(strdup(search_path));
+	new->len = len;
 	new->next = search;
 
 	return new;
@@ -1051,16 +1014,17 @@ static struct module_overrides *add_override(const char *modfile,
 	
 }
 
-/* Recursion */
-static int read_config(const char *filename,
-		       const char *basedir,
-		       struct module_search **search,
-		       struct module_overrides **overrides);
+static int parse_config_scan(const char *filename,
+			     const char *basedir,
+			     const char *kernelversion,
+			     struct module_search **search,
+			     struct module_overrides **overrides);
 
-static int read_config_file(const char *filename,
-			    const char *basedir,
-			    struct module_search **search,
-                            struct module_overrides **overrides)
+static int parse_config_file(const char *filename,
+			     const char *basedir,
+			     const char *kernelversion,
+			     struct module_search **search,
+			     struct module_overrides **overrides)
 {
 	char *line;
 	unsigned int linenum = 0;
@@ -1085,41 +1049,75 @@ static int read_config_file(const char *filename,
 			continue;
 		}
 
-		if (strcmp(cmd, "search") == 0) {
+		if (streq(cmd, "search")) {
 			char *search_path;
 			
-			while ((search_path = strsep_skipspace(&ptr, "\t ")))
-				*search = add_search(search_path, *search);
-		} else if (strcmp(cmd, "override") == 0) {
+			while ((search_path = strsep_skipspace(&ptr, "\t "))) {
+				char *dirname;
+				size_t len;
+
+				if (strcmp(search_path,
+						MODULE_BUILTIN_KEY) == 0) {
+					*search = add_search(MODULE_BUILTIN_KEY,
+							     0, *search);
+					continue;
+				}
+				nofail_asprintf(&dirname, "%s%s%s/%s", basedir,
+					MODULE_DIR, kernelversion, search_path);
+				len = strlen(dirname);
+				*search = add_search(dirname, len, *search);
+				free(dirname);
+			}
+		} else if (streq(cmd, "override")) {
 			char *pathname = NULL, *version, *subdir;
 			modname = strsep_skipspace(&ptr, "\t ");
 			version = strsep_skipspace(&ptr, "\t ");
 			subdir = strsep_skipspace(&ptr, "\t ");
 
-			pathname = NOFAIL(malloc(strlen(basedir)
-				               + strlen(MODULE_DIR)
-					       + strlen(version)
-					       + strlen(subdir)
-					       + strlen(modname)
-					       + strlen(".ko")
-					       + 3));
-			sprintf(pathname, "%s%s%s/%s/%s.ko", basedir,
-				MODULE_DIR, version, subdir, modname);
+			if (strcmp(version, kernelversion) != 0 &&
+			    strcmp(version, "*") != 0)
+				continue;
+
+			nofail_asprintf(&pathname, "%s%s%s/%s/%s.ko", basedir,
+				MODULE_DIR, kernelversion, subdir, modname);
 
 			*overrides = add_override(pathname, *overrides);
-		} else if (strcmp(cmd, "include") == 0) {
+			free(pathname);
+		} else if (streq(cmd, "include")) {
 			char *newfilename;
 
 			newfilename = strsep_skipspace(&ptr, "\t ");
-			if (!newfilename)
+			if (!newfilename) {
 				grammar(cmd, filename, linenum);
-			else {
-				if (!read_config(newfilename, basedir,
-						 search, overrides))
+			} else {
+				warn("\"include %s\" is deprecated, "
+				     "please use /etc/depmod.d\n", newfilename);
+				if (strstarts(newfilename, "/etc/depmod.d")) {
+					warn("\"include /etc/depmod.d\" is "
+					     "the default, ignored\n");
+				} else {
+					if (!parse_config_scan(newfilename, basedir,
+							       kernelversion,
+							       search, overrides))
 					warn("Failed to open included"
 					     " config file %s: %s\n",
 					     newfilename, strerror(errno));
-                        }
+				}
+			}
+		} else if (streq(cmd, "make_map_files")) {
+			char *option;
+
+			option = strsep_skipspace(&ptr, "\t ");
+			if (!option)
+				grammar(cmd, filename, linenum);
+			else {
+				if (streq(option, "yes"))
+					make_map_files = 1;
+				else if (streq(option, "no"))
+					make_map_files = 0;
+				else
+					grammar(cmd, filename, linenum);
+			}
                 } else
                         grammar(cmd, filename, linenum);
 
@@ -1129,66 +1127,97 @@ static int read_config_file(const char *filename,
         return 1;
 }
 
-/* Simple format, ignore lines starting with #, one command per line.
-   Returns true or false. */
-static int read_config(const char *filename,
-		       const char *basedir,
-		       struct module_search **search,
-                       struct module_overrides **overrides)
+static int parse_config_scan(const char *filename,
+			     const char *basedir,
+			     const char *kernelversion,
+			     struct module_search **search,
+			     struct module_overrides **overrides)
 {
 	DIR *dir;
 	int ret = 0;
 
 	dir = opendir(filename);
 	if (dir) {
+		struct file_entry {
+			struct list_head node;
+			char name[];
+		};
+		LIST_HEAD(files_list);
+		struct file_entry *fe, *fe_tmp;
 		struct dirent *i;
-		while ((i = readdir(dir)) != NULL) {
-			if (!streq(i->d_name,".") && !streq(i->d_name,"..")) {
-				char sub[strlen(filename) + 1
-				       + strlen(i->d_name) + 1];
 
-				sprintf(sub, "%s/%s", filename, i->d_name);
-				if (!read_config(sub, basedir, search,
-						 overrides))
-					warn("Failed to open"
-					     " config file %s: %s\n",
-					     sub, strerror(errno));
-			}
+		/* sort files from directory into list */
+		while ((i = readdir(dir)) != NULL) {
+			size_t len;
+
+			if (i->d_name[0] == '.')
+				continue;
+			if (!config_filter(i->d_name))
+				continue;
+
+			len = strlen(i->d_name);
+			if (len < 6 || strcmp(&i->d_name[len-5], ".conf") != 0)
+				warn("All config files need .conf: %s/%s, "
+				     "it will be ignored in a future release.\n",
+				     filename, i->d_name);
+			fe = malloc(sizeof(struct file_entry) + len + 1);
+			if (fe == NULL)
+				continue;
+			strcpy(fe->name, i->d_name);
+			list_for_each_entry(fe_tmp, &files_list, node)
+				if (strcmp(fe_tmp->name, fe->name) >= 0)
+					break;
+			list_add_tail(&fe->node, &fe_tmp->node);
 		}
 		closedir(dir);
+
+		/* parse list of files */
+		list_for_each_entry_safe(fe, fe_tmp, &files_list, node) {
+			char *cfgfile;
+
+			nofail_asprintf(&cfgfile, "%s/%s", filename, fe->name);
+			if (!parse_config_file(cfgfile, basedir, kernelversion,
+					       search, overrides))
+				warn("Failed to open config file "
+				     "%s: %s\n", fe->name, strerror(errno));
+			free(cfgfile);
+			list_del(&fe->node);
+			free(fe);
+		}
+
 		ret = 1;
 	} else {
-		if (read_config_file(filename, basedir, search, overrides))
+		if (parse_config_file(filename, basedir, kernelversion, search,
+				      overrides))
 			ret = 1;
 	}
 
 	return ret;
 }
 
-static const char *default_configs[] =
+static void parse_toplevel_config(const char *filename,
+				  const char *basedir,
+				  const char *kernelversion,
+				  struct module_search **search,
+				  struct module_overrides **overrides)
 {
-	"/etc/depmod.conf",
-	"/etc/depmod.d",
-};
-
-static void read_toplevel_config(const char *filename,
-				 const char *basedir,
-				 struct module_search **search,
-                                 struct module_overrides **overrides)
-{
-	unsigned int i;
-
 	if (filename) {
-		if (!read_config(filename, basedir, search, overrides))
+		if (!parse_config_scan(filename, basedir, kernelversion, search,
+				 overrides))
 			fatal("Failed to open config file %s: %s\n",
 			      filename, strerror(errno));
 		return;
 	}
 
-	/* Try defaults. */
-	for (i = 0; i < ARRAY_SIZE(default_configs); i++) {
-		read_config(default_configs[i], basedir, search, overrides);
-	}
+	/* deprecated config file */
+	if (parse_config_file("/etc/depmod.conf", basedir, kernelversion,
+			      search, overrides) > 0)
+		warn("Deprecated config file /etc/depmod.conf, "
+		      "all config files belong into /etc/depmod.d/.\n");
+
+	/* default config */
+	parse_config_scan("/etc/depmod.d", basedir, kernelversion,
+			  search, overrides);
 }
 
 /* Local to main, but not freed on exit.  Keep valgrind quiet. */
@@ -1199,26 +1228,33 @@ struct module_overrides *overrides = NULL;
 int main(int argc, char *argv[])
 {
 	int opt, all = 0, maybe_all = 0, doing_stdout = 0;
-	char *basedir = "", *dirname, *version, *badopt = NULL,
-		*system_map = NULL;
+	char *basedir = "", *dirname, *version;
+	char *system_map = NULL, *module_symvers = NULL;
 	int i;
 	const char *config = NULL;
 
-	/* Don't print out any errors just yet, we might want to exec
-           backwards compat version. */
-	opterr = 0;
-	while ((opt = getopt_long(argc, argv, "ab:ArehnqruvVF:C:w", options, NULL))
+	if (native_endianness() == 0)
+		abort();
+
+	while ((opt = getopt_long(argc, argv, "aAb:C:E:F:euqrvnhVwm", options, NULL))
 	       != -1) {
 		switch (opt) {
 		case 'a':
 			all = 1;
 			break;
+		case 'A':
+			maybe_all = 1;
+			break;
 		case 'b':
 			basedir = optarg;
 			skipchars = strlen(basedir);
 			break;
-		case 'A':
-			maybe_all = 1;
+		case 'C':
+			config = optarg;
+			break;
+		case 'E':
+			module_symvers = optarg;
+			check_symvers = 1;
 			break;
 		case 'F':
 			system_map = optarg;
@@ -1226,39 +1262,43 @@ int main(int argc, char *argv[])
 		case 'e':
 			print_unknown = 1;
 			break;
-		case 'v':
-			verbose = 1;
-			break;
 		case 'u':
 		case 'q':
 		case 'r':
 			break;
-		case 'C':
-			config = optarg;
+		case 'v':
+			verbose = 1;
+			break;
+		case 'n':
+			doing_stdout = 1;
 			break;
 		case 'h':
 			print_usage(argv[0]);
 			exit(0);
 			break;
-		case 'n':
-			doing_stdout = 1;
-			break;
 		case 'V':
 			printf("%s %s\n", PACKAGE, VERSION);
 			exit(0);
 		case 'w':
 			warn_dups = 1;
 			break;
+		case 'm':
+			force_map_files = 1;
+			break;
 		default:
-			badopt = argv[optind-1];
+			print_usage(argv[0]);
+			exit(1);
 		}
 	}
 
-	/* We can't print unknowns without a System.map */
-	if (!system_map)
-		print_unknown = 0;
-	else
+	if (module_symvers)
+		load_module_symvers(module_symvers);
+	else if (system_map)
 		load_system_map(system_map);
+	else if (print_unknown) {
+		warn("-e needs -E or -F");
+		print_unknown = 0;
+	}
 
 	/* They can specify the version naked on the command line */
 	if (optind < argc && is_version_number(argv[optind])) {
@@ -1270,25 +1310,18 @@ int main(int argc, char *argv[])
 		version = NOFAIL(strdup(buf.release));
 	}
 
-	/* Run old version if required. */
-	if (old_module_version(version))
-		exec_old_depmod(argv);
-
-	if (badopt) {
-		fprintf(stderr, "%s: malformed/unrecognized option '%s'\n",
-			argv[0], badopt);
-		print_usage(argv[0]);
-		exit(1);
+	/* Check for old version. */
+	if (old_module_version(version)) {
+		fprintf(stderr, "Kernel version %s requires old depmod\n",
+			version);
+		exit(2);
 	}
 
 	/* Depmod -a by default if no names. */
 	if (optind == argc)
 		all = 1;
 
-	dirname = NOFAIL(malloc(strlen(basedir)
-			 + strlen(MODULE_DIR)
-			 + strlen(version) + 1));
-	sprintf(dirname, "%s%s%s", basedir, MODULE_DIR, version);
+	nofail_asprintf(&dirname, "%s%s%s", basedir, MODULE_DIR, version);
 
 	if (maybe_all) {
 		if (!doing_stdout && !depfile_out_of_date(dirname))
@@ -1296,27 +1329,41 @@ int main(int argc, char *argv[])
 		all = 1;
 	}
 
-	read_toplevel_config(config, basedir, &search, &overrides);
+	parse_toplevel_config(config, basedir, version, &search, &overrides);
 
 	/* For backward compatibility add "updates" to the head of the search
 	 * list here. But only if there was no "search" option specified.
 	 */
+	if (!search) {
+		char *dirname;
+		size_t len;
 
-	if (!search)
-		search = add_search("updates",search);
-	
+		nofail_asprintf(&dirname, "%s%s%s/updates", basedir,
+				MODULE_DIR, version);
+		len = strlen(dirname);
+		search = add_search(dirname, len, search);
+	}
 	if (!all) {
 		/* Do command line args. */
 		for (opt = optind; opt < argc; opt++) {
-			struct module *new = grab_module(NULL, argv[opt]);
-			if (new) {
-				new->next = list;
-				list = new;
+			struct module *new;
+
+			if (argv[opt][0] != '/')
+				fatal("modules must be specified using absolute paths.\n"
+					"\"%s\" is a relative path\n", argv[opt]);
+
+			new = grab_module(NULL, argv[opt]);
+			if (!new) {
+				/* cmd-line specified modules must exist */
+				fatal("grab_module() failed for module %s\n", argv[opt]);
 			}
+			new->next = list;
+			list = new;
 		}
 	} else {
 		list = grab_basedir(dirname,search,overrides);
 	}
+	list = sort_modules(dirname,list);
 	list = parse_modules(list);
 
 	for (i = 0; i < sizeof(depfiles)/sizeof(depfiles[0]); i++) {
@@ -1326,6 +1373,9 @@ int main(int argc, char *argv[])
 		char tmpname[strlen(dirname) + 1 + strlen(d->name) +
 						strlen(".temp") + 1];
 
+		if (d->map_file && !make_map_files && !force_map_files)
+			continue;
+
 		sprintf(depname, "%s/%s", dirname, d->name);
 		sprintf(tmpname, "%s/%s.temp", dirname, d->name);
 		if (!doing_stdout) {
@@ -1338,7 +1388,7 @@ int main(int argc, char *argv[])
 			if (ends_in(depname, ".bin"))
 				continue;
 		}
-		d->func(list, out);
+		d->func(list, out, dirname);
 		if (!doing_stdout) {
 			fclose(out);
 			if (rename(tmpname, depname) < 0)
